
import ctypes
import infi.wioctl
import infi.cwrap
from . import constants
from . import structures

# pylint: disable=W0622

def _sizeof(struct):
    return struct.min_max_sizeof().max

def _extract_whole_structure_from_drive_layout_buffer(string):
    instance = structures.DRIVE_LAYOUT_INFORMATION_EX.create_from_string(string)
    return instance


class CoCreateGuid(infi.wioctl.api.WrappedFunction):
    return_value = infi.wioctl.api.BOOL

    @classmethod
    def get_errcheck(cls):
        return infi.cwrap.errcheck_nonzero()

    @classmethod
    def get_library_name(cls):
        return 'ole32'

    @classmethod
    def get_parameters(cls):
        return ((infi.wioctl.api.c_void_p, infi.wioctl.api.IN_OUT, "guid"),
                )

def generate_guid():
    size = _sizeof(infi.wioctl.structures.GUID)
    buffer = ctypes.c_buffer('\x00' * size)
    CoCreateGuid(buffer)
    return infi.wioctl.structures.GUID.create_from_string(buffer)

def generate_signature():
    import random
    return random.randint(0, 2 ** 32 - 1)

GUID_ZERO = infi.wioctl.structures.GUID.create_from_string(ctypes.c_buffer('\x00' * 16))

class DeviceIoControl(infi.wioctl.DeviceIoControl):
    def _partial_ioctl_diks_get_drive_layout_ex(self, size=24):
        """:returns: a `ctypes.c_buffer` object"""
        # http://msdn.microsoft.com/en-us/library/windows/hardware/ff560364(v=VS.85).aspx
        # To determine the size of output buffer that is required,
        # caller should send this IOCTL request in a loop.
        # Every time the storage stack rejects the IOCTL with an error message indicating that the buffer was too small,
        # caller should double the buffer size.
        buffer = ctypes.c_buffer('\x00' * size, size)
        try:
            self.ioctl(infi.wioctl.constants.IOCTL_DISK_GET_DRIVE_LAYOUT_EX, 0, 0, buffer, size)
        except infi.wioctl.api.WindowsException, e:
            if e.winerror != infi.wioctl.constants.ERROR_INSUFFICIENT_BUFFER:
                raise
            return self._partial_ioctl_diks_get_drive_layout_ex(size * 2)
        return buffer

    def ioctl_disk_get_drive_layout_ex(self):
        buffer = self._partial_ioctl_diks_get_drive_layout_ex()
        return _extract_whole_structure_from_drive_layout_buffer(buffer)

    def ioctl_disk_set_drive_layout_ex(self, layout):
        size = layout.sizeof(layout)
        input_buffer = ctypes.c_buffer(layout.write_to_string(layout))
        output_buffer = ctypes.c_buffer('\x00' * size, size)
        self.ioctl(infi.wioctl.constants.IOCTL_DISK_SET_DRIVE_LAYOUT_EX, input_buffer, size, output_buffer, size)

    def ioctl_disk_delete_drive_layout(self):
        self.ioctl(infi.wioctl.constants.IOCTL_DISK_DELETE_DRIVE_LAYOUT, 0, 0, 0, 0)

    def ioctl_disk_create_disk(self, partition_style):
        """:param partition_style: int
        :param signature: int
        :param disk_id: 16-byte string
        :param max_partition_count: int"""
        struct = structures.CREATE_DISK()
        struct.PartitionStyle = partition_style
        if struct.PartitionStyle == constants.PARTITION_STYLE_MBR:
            klass = structures.CREATE_DISK_MBR
            union = klass()
            union.Signature = generate_signature()
        else:
            klass = structures.CREATE_DISK_GPT
            union = klass()
            union.DiskId = generate_guid()
            union.MaxPartitionCount = 128
        struct.union = union
        size = _sizeof(struct)
        buffer = ctypes.c_buffer(structures.CREATE_DISK.write_to_string(struct), size)
        self.ioctl(infi.wioctl.constants.IOCTL_DISK_CREATE_DISK, buffer, size, 0, 0)

    def _bool_ioctl(self, ioctl_number):
        try:
            self.ioctl(ioctl_number, 0, 0, 0, 0)
        except infi.wioctl.api.WindowsException, e:
            if e.winerror != 1:
                raise
            return False
        return True

    def ioctl_volume_is_partition(self):
        return self._bool_ioctl(infi.wioctl.constants.IOCTL_VOLUME_IS_PARTITION)

    def ioctl_volume_is_offline(self):
        return self._bool_ioctl(infi.wioctl.constants.IOCTL_VOLUME_IS_OFFLINE)

    def ioctl_volume_supports_online_offline(self):
        return self._bool_ioctl(infi.wioctl.constants.IOCTL_VOLUME_SUPPORTS_ONLINE_OFFLINE)

    def ioctl_volume_is_io_capable(self):
        return self._bool_ioctl(infi.wioctl.constants.IOCTL_VOLUME_IS_IO_CAPABLE)

    def ioctl_volume_online(self):
        return self._bool_ioctl(infi.wioctl.constants.IOCTL_VOLUME_ONLINE)

    def ioctl_volume_offline(self):
        return self._bool_ioctl(infi.wioctl.constants.IOCTL_VOLUME_OFFLINE)

    def ioctl_disk_get_disk_attributes(self):
        klass = structures.GET_DISK_ATTRIBUTES
        size = _sizeof(klass)
        buffer = ctypes.c_buffer('\x00' * size, size)
        self.ioctl(infi.wioctl.constants.IOCTL_DISK_GET_DISK_ATTRIBUTES, 0, 0, buffer, size)
        return klass.create_from_string(buffer)

    def ioctl_disk_set_disk_attributes(self, attributes):
        size = attributes.sizeof(attributes)
        input_buffer = ctypes.c_buffer(attributes.write_to_string(attributes))
        self.ioctl(infi.wioctl.constants.IOCTL_DISK_SET_DISK_ATTRIBUTES, input_buffer, size, 0, 0)

    def ioctl_disk_get_san_settings(self):
        klass = structures.DISK_SAN_SETTINGS
        size = _sizeof(klass)
        buffer = ctypes.c_buffer('\x00' * size, size)
        self.ioctl(infi.wioctl.constants.IOCTL_DISK_GET_SAN_SETTINGS, 0, 0, buffer, size)
        return klass.create_from_string(buffer)

    def ioctl_disk_set_san_settings(self, settings):
        klass = structures.DISK_SAN_SETTINGS
        size = _sizeof(klass)
        buffer = ctypes.c_buffer(settings.write_to_string(settings))
        self.ioctl(infi.wioctl.constants.IOCTL_DISK_SET_SAN_SETTINGS, buffer, size, 0, 0)

    def ioctl_volume_query_volume_number(self):
        klass = structures.VOLUME_NUMBER
        size = _sizeof(klass)
        buffer = ctypes.c_buffer('\x00' * size, size)
        self.ioctl(infi.wioctl.constants.IOCTL_VOLUME_QUERY_VOLUME_NUMBER, 0, 0, buffer, size)
        return klass.create_from_string(buffer)

    def _partial_ioctl_volume_get_volume_disk_extents(self, size):
        klass = structures.VOLUME_DISK_EXTENTS
        buffer = ctypes.c_buffer('\x00' * size, size)
        try:
            self.ioctl(infi.wioctl.constants.IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS, 0, 0, buffer, size)
        except infi.wioctl.api.WindowsException, e:
            if e.winerror != infi.wioctl.constants.ERROR_MORE_DATA:
                raise
            return self._partial_ioctl_volume_get_volume_disk_extents(size + _sizeof(structures.DISK_EXTENT))
        return klass.create_from_string(buffer)

    def ioctl_volume_get_volume_disk_extents(self):
        size = _sizeof(structures.DISK_EXTENT) + ctypes.sizeof(ctypes.c_ulong) + 4
        return self._partial_ioctl_volume_get_volume_disk_extents(size)

    def _partial_ioctl_mountmgr_query_points(self, input_buffer, input_buffer_size, size=256):
        """:returns: a `ctypes.c_buffer` object"""
        buffer = ctypes.c_buffer('\x00' * size, size)
        try:
            self.ioctl(infi.wioctl.constants.IOCTL_MOUNTMGR_QUERY_POINTS, input_buffer, input_buffer_size, buffer, size)
        except infi.wioctl.api.WindowsException, e:
            if e.winerror != infi.wioctl.constants.ERROR_MORE_DATA:
                raise
            return self._partial_ioctl_mountmgr_query_points(input_buffer, input_buffer_size, size * 2)
        return buffer

    def ioctl_mountmgr_query_points(self, input_buffer, input_buffer_size):
        return self._partial_ioctl_mountmgr_query_points(input_buffer, input_buffer_size)

    def ioctl_mountmgr_query_auto_mount(self):
        klass = structures.MOUNTMGR_QUERY_AUTO_MOUNT
        size = _sizeof(klass)
        buffer = ctypes.c_buffer('\x00' * size, size)
        self.ioctl(infi.wioctl.constants.IOCTL_MOUNTMGR_QUERY_AUTO_MOUNT, 0, 0, buffer, size)
        return klass.create_from_string(buffer).CurrentState

    def ioctl_mountmgr_set_auto_mount(self, state):
        klass = structures.MOUNTMGR_SET_AUTO_MOUNT
        size = _sizeof(klass)
        buffer = ctypes.c_buffer(klass.write_to_string(klass(NewState=state)))
        self.ioctl(infi.wioctl.constants.IOCTL_MOUNTMGR_SET_AUTO_MOUNT, buffer, size, 0, 0)

    def ioctl_disk_grow_partition(self, grow_struct):
        size = grow_struct.sizeof(grow_struct)
        input_buffer = ctypes.c_buffer(structures.DISK_GROW_PARTITION.write_to_string(grow_struct))
        self.ioctl(infi.wioctl.constants.IOCTL_DISK_GROW_PARTITION, input_buffer, size, 0, 0)

    def ioctl_extend_volume(self, new_size):
        size_in_sectors = ctypes.c_longlong(new_size / 512)
        self.ioctl(infi.wioctl.constants.FSCTL_EXTEND_VOLUME,
                   ctypes.byref(size_in_sectors),
                   ctypes.sizeof(size_in_sectors), 0, 0)

    def ioctl_disk_update_drive_size(self):
        size = _sizeof(structures.DISK_GEOMETRY)
        buffer = ctypes.c_buffer('\x00' * size, size)
        try:
            self.ioctl(infi.wioctl.constants.IOCTL_DISK_UPDATE_DRIVE_SIZE, 0, 0, buffer, size)
        except infi.wioctl.api.WindowsException, e:
            if e.winerror == infi.wioctl.constants.ERROR_INSUFFICIENT_BUFFER:
                return
            raise

    def ioctl_disk_update_properties(self):
        self.ioctl(infi.wioctl.constants.IOCTL_DISK_UPDATE_PROPERTIES, 0, 0, 0, 0)





from capacity import byte, MiB
from infi.pyutils.lazy import cached_method, clear_cache
from infi.pyutils.decorators import wraps
from ..ioctl import DeviceIoControl, generate_signature, generate_guid, structures, constants
from ..ioctl.constants import PARTITION_STYLE_MBR, PARTITION_STYLE_GPT, PARTITION_STYLE_RAW
from infi.diskmanagement.ioctl import GUID_ZERO
from infi.wioctl.structures import is_64bit, GUID
from ..mount_manager import MountManager
from infi.devicemanager import DeviceManager
from logging import getLogger

logger = getLogger(__name__)

def is_zero(large_integer):
    """:returns: if the value of LARGE_INTEGER is zero"""
    if len(large_integer._fields_) == 1:
        return large_integer.QuadPart == 0
    return large_integer.LowPart == 0 and large_integer.HighPart == 0

# http://msdn.microsoft.com/en-us/windows/hardware/gg463525
PARTITION_BASIC_DATA_GUID = GUID(Data1=0xEBD0A0A2L, Data2=0xB9E5, Data3=0x4433,
                                 Data4=[0x87, 0xC0, 0x68, 0xB6, 0xB7, 0x26, 0x99, 0xC7])
PARTITION_MSFT_RESERVED_GUID = GUID(Data1=0xE3C9E316L, Data2=0x0B5C, Data3=0x4DB8,
                                    Data4=[0x81, 0x7D, 0xF9, 0x2D, 0xF0, 0x02, 0x15, 0xAE])
PARTITION_MSFT_RESERVED_STARTING_OFFSET = 17408
PARTITION_MSFT_RESERVED_SIZE_MIN = 32 * 1024 * 1024
PARTITION_MSFT_RESERVED_SIZE_MAX = 128 * 1024 * 1024
PARTITION_MSFT_RESERVED_BAR = 16 * 1024 * 1024 * 1024
ERROR_INVALID_PARAMETER = 87
FIRST_PRIMARY_PARTITION_OFFSET = 1024 * 1024
GPT_PARTITION_OFFSET = 1024 * 1024

# Windows returns empty partition entries from the partition table.
# This lambda-function that returns True if the partition is not empty, i.e. in use
partitions_in_use_lambda = lambda partition: not is_zero(partition.PartitionLength)


def is_guid_partition(partition):
    return partition.union.PartitionType.Data1 == PARTITION_MSFT_RESERVED_GUID.Data1 and \
           partition.union.PartitionType.Data2 == PARTITION_MSFT_RESERVED_GUID.Data2 and \
           partition.union.PartitionType.Data3 == PARTITION_MSFT_RESERVED_GUID.Data3 and \
           partition.union.PartitionType.Data4 == PARTITION_MSFT_RESERVED_GUID.Data4


def to_large_integer(number):
    kwargs = {}
    if is_64bit():
        kwargs['QuadPart'] = number
    else:
        kwargs['HighPart'] = (number & 0xFFFFFFFF00000000) >> 32
        kwargs['LowPart'] = number & 0xFFFFFFFF
    instance = structures.LARGE_INTEGER(**kwargs)
    return instance

def from_large_integer(instance):
    return instance.QuadPart if is_64bit() else ((instance.HighPart << 32) + instance.LowPart)

def partition_type_specific(func):
    @wraps(func)
    def callee(*args, **kwargs):
        _name = func.func_name
        _self, args = args[0], args[1:]
        _type = _self._get_named_type()
        return getattr(_self, '_{}_{}'.format(_name, _type))(*args, **kwargs)
    return callee

class Volume(object):
    def __init__(self, disk, partition, mount_manager=None):
        super(Volume, self).__init__()
        self._disk = disk
        self._partition = partition
        self._mount_manager = mount_manager or MountManager()
        self._io = DeviceIoControl(self._path, False)

    @property
    @cached_method
    def _path(self):
        return self.get_volume_guid()

    def __repr__(self):
        try:
            return "Volume <{}>".format(self._path)
        except:
            return "<Volume <?>"

    @classmethod
    def get_from_disk_and_partition(cls, disk, partition, mount_manager=None):
        return Volume(disk, partition, mount_manager=mount_manager)

    def _get_device_number(self):
        from infi.devicemanager.ioctl import DeviceIoControl
        return DeviceIoControl(self._path).storage_get_device_and_partition_number()

    @cached_method
    def _get_wmi_object(self):
        from ..wmi import WmiClient, iter_volumes
        from infi.devicemanager.ioctl import DeviceIoControl
        client = WmiClient()
        expected = self._get_device_number()
        def _filter(volume):
            actual = DeviceIoControl(volume.DeviceID.rstrip(r'\\')).storage_get_device_and_partition_number()
            return actual == expected
        return filter(_filter, iter_volumes(client))[0]

    def format(self, quick=True, file_system="NTFS"):
        # TODO the idea is to do only the formatting through wmi
        # next step is to figure out to quickly get from the setuapi and ioctl information to the wmi object
        self.online()
        wmi_object = self._get_wmi_object()
        wmi_object.Format(QuickFormat=quick, FileSystem=file_system)

    def get_volume_guid(self):
        from infi.wioctl.errors import WindowsException

        _struct = self._partition._struct
        partition_info = from_large_integer(_struct.StartingOffset), from_large_integer(_struct.PartitionLength)

        for volume_guid in self._mount_manager.get_volume_guids():
            volume_guid = volume_guid.rstrip("\\")
            try:
                extents = self._mount_manager.get_volume_extents(volume_guid)
            except WindowsException,e:
                if e.winerror == 1: # Floppy/CD/..
                    continue
                raise
            for extent in extents:
                extent_info = from_large_integer(extent.StartingUsableOffset), from_large_integer(extent.ExtentLength)
                if extent.DiskNumber != self._disk._number:
                    continue
                # extent needs to be contained inside the partition
                if extent_info[0] >= partition_info[0] and sum(extent_info) <= sum(partition_info):
                    return volume_guid

    def get_moint_points(self):
        return self._mount_manager.get_volume_mount_points(self)

    def add_mount_point(self, mount_point):
        return self._mount_manager.add_volume_mount_point(self.get_volume_guid(), mount_point)

    def remove_mount_point(self, mount_point):
        return self._mount_manager.remove_volume_mount_point(mount_point)

    def has_drive_letter(self):
        return self._mount_manager.get_volume_drive_letter(self) is not None

    def get_available_drive_letters(self):
        return self._mount_manager.get_avaialable_drive_letters()

    def assign_first_available_drive_letter(self):
        return self.add_mount_point(self.get_available_drive_letters()[0])

    def online(self):
        return DeviceIoControl(self._path, True, False).ioctl_volume_online()

    def offline(self):
        return DeviceIoControl(self._path, True, False).ioctl_volume_offline()

    def resize(self, size_in_bytes):
        return DeviceIoControl(self._path, True, False).ioctl_extend_volume(size_in_bytes)


class Partition(object):
    def __init__(self, disk, struct):
        super(Partition, self).__init__()
        self._struct = struct
        self._disk = disk

    def __repr__(self):
        start, size = byte * self.get_start_offset_in_bytes(), byte * self.get_size_in_bytes()
        return "Partition <start={}, capacity={}> on {!r}".format(start, size, self._disk)

    def is_gpt(self):
        return self._get_layout().PartitionStyle == PARTITION_STYLE_GPT

    def is_mbr(self):
        return self._get_layout().PartitionStyle == PARTITION_STYLE_MBR

    def is_raw(self):
        return self._get_layout().PartitionStyle == PARTITION_STYLE_RAW

    def is_empty(self):
        return is_zero(self._struct.PartitionLength)

    def get_size_in_bytes(self):
        return from_large_integer(self._struct.PartitionLength)

    def get_start_offset_in_bytes(self):
        return from_large_integer(self._struct.StartingOffset)

    @classmethod
    def _create(cls, disk, index, number, start_offset, length):
        partition = structures.PARTITION_INFORMATION_EX()
        layout = disk._get_layout()
        partition.StartingOffset = start_offset
        partition.PartitionStyle = layout.PartitionStyle
        partition.PartitionLength = length
        partition.PartitionNumber = number
        partition.RewritePartition = 1
        if partition.PartitionStyle == constants.PARTITION_STYLE_GPT:
            partition.union = structures.PARTITION_INFORMATION_GPT()
        else:
            partition.union = structures.PARTITION_INFORMATION_MBR()
        try:
            layout.PartitionEntry[index] = partition
        except IndexError:
            layout.PartitionEntry = [p for p in layout.PartitionEntry] + [partition, ]
        layout.PartitionCount = len(layout.PartitionEntry)
        return partition

    @classmethod
    def create_primary(cls, disk, index=0, boot=False, start_offset_in_bytes=FIRST_PRIMARY_PARTITION_OFFSET, size_in_bytes=None):
        """:param size: if size_in_bytes is None, the partition will be for the entire disk
        :param offset: either a number or Capacity
        :param size_in_bytes: """
        # I did not get this information from any official documentation, just by doing what the Disk Management does
        partition = cls._create(disk, index, index + 1,
                                to_large_integer(start_offset_in_bytes),
                                to_large_integer(disk.get_size_in_bytes() - 4 * 1024 * 1024 \
                                                 if size_in_bytes is None else size_in_bytes))
        partition.union.PartitionType = constants.PARTITION_HUGE
        partition.union.BootIndicator = constants.TRUE if boot else 0
        partition.union.RecognizedPartition = constants.TRUE
        partition.union.HiddenSectors = 0
        disk._update_layout()

    @classmethod
    def create_guid(cls, disk, index, partition_type_guid, start_offset_in_bytes, size_in_bytes):
        """:param size: if size_in_bytes is None, the partition will be for the entire disk
        :param offset: either a number or Capacity
        :param size_in_bytes: """
        # I did not get this information from any official documentation, just by doing what the Disk Management
        partition = cls._create(disk, index, index,
                                to_large_integer(start_offset_in_bytes),
                                to_large_integer(disk.get_size_in_bytes() - 65 * 1024 * 1024 \
                                                 if size_in_bytes is None else size_in_bytes))
        partition.union.PartitionType = partition_type_guid
        partition.union.PartitionId = generate_guid()
        partition.union.Attributes = 0
        partition.union.Name = [0, ]*36
        disk._update_layout()

    @cached_method
    def get_volume(self):
        return Volume.get_from_disk_and_partition(self._disk, self)

    def resize(self, size_in_bytes):
        import infi.wioctl
        self._disk._io.ioctl_disk_update_properties()
        self._disk._io.ioctl_disk_update_drive_size()

        _struct = structures.DISK_GROW_PARTITION()
        _struct.PartitionNumber = self._struct.PartitionNumber
        _struct.BytesToGrow = to_large_integer(size_in_bytes - self.get_size_in_bytes())
        try:
            self._disk._io.ioctl_disk_grow_partition(_struct)
        except infi.wioctl.api.WindowsException, e:
            if e.winerror != ERROR_INVALID_PARAMETER:
                raise
            _struct.BytesToGrow = to_large_integer(size_in_bytes - self.get_size_in_bytes() - 32000)
            self._disk._io.ioctl_disk_grow_partition(_struct)
        self._struct.PartitionLength = to_large_integer(size_in_bytes)  # we get_size_in_bytes will return the new size


class Disk(object):
    def __init__(self, disk_number):
        super(Disk, self).__init__()
        self._number = disk_number
        self._path = r"\\.\PHYSICALDRIVE{}".format(self._number)
        self._io = DeviceIoControl(self._path, True)

    def __repr__(self):
        return "Disk <{}>".format(self._path)

    @cached_method
    def _get_layout(self):
        """:returns: a DRIVE_LAYOUT_INFORMATION_EX Structure"""
        return self._io.ioctl_disk_get_drive_layout_ex()

    def _set_layout(self, layout):
        self._io.ioctl_disk_set_drive_layout_ex(layout)
        self.wait_for_all_volumes()
        self.clear_cached_properties()

    def wait_for_all_volumes(self):
        from time import sleep
        from waiting import wait
        def predicate():
            try:
                _ = [partition.get_volume().get_volume_guid() for partition in self.get_partitions()]
                return True
            except:
                logger.exception("Exception in wait_for_all_volumes")
                return False
        wait(predicate, timeout_seconds=30)

    def _update_layout(self):
        self._set_layout(self._get_layout())

    @cached_method
    def get_size_in_bytes(self):
        import infi.devicemanager.ioctl
        return infi.devicemanager.ioctl.DeviceIoControl(self._path).disk_get_drive_geometry_ex()

    def _get_named_type(self):
        if self.is_gpt():
            return 'gpt'
        elif self.is_mbr():
            return 'mbr'
        else:
            return 'raw'

    def is_gpt(self):
        return self._get_layout().PartitionStyle == PARTITION_STYLE_GPT

    def is_mbr(self):
        return self._get_layout().PartitionStyle == PARTITION_STYLE_MBR

    def is_raw(self):
        return self._get_layout().PartitionStyle == PARTITION_STYLE_RAW

    def clear_cached_properties(self):
        clear_cache(self)

    def _has_extended_partition(self):
        return len(self._get_layout().PartitionEntry) >= 3 and \
               not is_zero(self._get_layout().PartitionEntry[3].PartitionLength)

    def _iter_partitions_mbr(self):
        for struct in self._get_layout().PartitionEntry[0:3]:
            # The first three MBR partitions are always primary (can be empty entries)
            partition = Partition(self, struct)
            if partition.is_empty():
                continue
            yield partition
        if not self._has_extended_partition():
            return
        for struct in self._get_layout().PartitionEntry[4::4]:
            partition = Partition(self, struct)
            if partition.is_empty():
                continue
            yield partition

    def _iter_partitions_raw(self):
        pass

    def _iter_partitions_gpt(self):
        # When creating a GPT drive in Windows, the first partition is a hidden partition
        # When creating a VSS snapshot, Windows moves the reseved partition and creates another one before it
        # This method needs to return the non-hidden partitions
        # The only logic I see that fits is to yield those after the reserved parititon:
        # the reserved partition has a known GUID, I can't find a known guid for the VSS partition
        key_lambda = lambda item: from_large_integer(item.StartingOffset)
        sorted_by_offset = sorted(self._get_layout().PartitionEntry, key=key_lambda)
        reserved_partition = [item for item in sorted_by_offset if is_guid_partition(item)]
        starting_index = 0
        if reserved_partition:
            starting_index = sorted_by_offset.index(reserved_partition[0])
        for struct in sorted_by_offset[starting_index+1:]:
            partition = Partition(self, struct)
            yield partition

    def get_partitions(self):
        return [partition for partition in self.iter_partitions()]

    def iter_partitions(self):
        for partition in getattr(self, "_iter_partitions_{}".format(self._get_named_type()))():
            yield partition

    def destroy_partition_table(self):
        self._io.ioctl_disk_delete_drive_layout()
        self.clear_cached_properties()

    def _create_partition_table_mbr(self, alignment_in_bytes=None):
        signature = generate_signature()
        self._io.ioctl_disk_create_disk(partition_style=PARTITION_STYLE_MBR)

    def _create_partition_table_gpt(self, alignment_in_bytes=None):
        guid = generate_guid()
        self._io.ioctl_disk_create_disk(partition_style=PARTITION_STYLE_GPT)
        minimum_required_size_in_bytes = PARTITION_MSFT_RESERVED_SIZE_MIN \
                                         if self.get_size_in_bytes() < PARTITION_MSFT_RESERVED_BAR \
                                         else PARTITION_MSFT_RESERVED_SIZE_MAX
        size_in_bytes = minimum_required_size_in_bytes
        if alignment_in_bytes:
            size_alignment = size_in_bytes % alignment_in_bytes
            offset_alignment = PARTITION_MSFT_RESERVED_STARTING_OFFSET % alignment_in_bytes
            if size_in_bytes % alignment_in_bytes:
                size_in_bytes += (alignment_in_bytes - size_in_bytes % alignment_in_bytes)
            if offset_alignment:
                size_in_bytes += (alignment_in_bytes - offset_alignment)
        Partition.create_guid(self, 1, PARTITION_MSFT_RESERVED_GUID,
                              PARTITION_MSFT_RESERVED_STARTING_OFFSET, size_in_bytes)

    def create_partition_table(self, type_name, alignment_in_bytes=None):
        """:param type_name: either 'gpt' or 'mbr'"""
        getattr(self, "_create_partition_table_{}".format(type_name))(alignment_in_bytes)
        self.clear_cached_properties()

    def create_first_partition(self, alignment_in_bytes=None):
        if self.is_mbr():
            offset_in_bytes = FIRST_PRIMARY_PARTITION_OFFSET
            if alignment_in_bytes:
                offset_alignment = FIRST_PRIMARY_PARTITION_OFFSET % alignment_in_bytes
                if offset_alignment:
                    offset_in_bytes += FIRST_PRIMARY_PARTITION_OFFSET - offset_alignment
            Partition.create_primary(self, start_offset_in_bytes=offset_in_bytes)
        elif self.is_gpt():
            reserved_partition = self._get_layout().PartitionEntry[0]
            offset_in_bytes = from_large_integer(reserved_partition.StartingOffset) + from_large_integer(reserved_partition.PartitionLength)
            offset_in_bytes += GPT_PARTITION_OFFSET
            if alignment_in_bytes:
                offset_alignment = GPT_PARTITION_OFFSET % alignment_in_bytes
                if offset_alignment:
                    offset_in_bytes += GPT_PARTITION_OFFSET - offset_alignment
            size_in_bytes = self.get_size_in_bytes() - offset_in_bytes - GPT_PARTITION_OFFSET
            Partition.create_guid(self, 2, PARTITION_BASIC_DATA_GUID, offset_in_bytes, size_in_bytes)

    def is_offline(self):
        return self._io.ioctl_disk_get_disk_attributes().Attributes & constants.DISK_ATTRIBUTE_OFFLINE

    def is_online(self):
        return not self.is_offline()

    def is_read_only(self):
        return self._io.ioctl_disk_get_disk_attributes().Attributes & constants.DISK_ATTRIBUTE_READ_ONLY

    def _set_disk_attrttributes(self, attributes, mask):
        set_struct = structures.SET_DISK_ATTRIBUTES(Version=0x28, Persist=constants.TRUE,
                                                    RelinquishOwnership=0, Attributes=attributes,
                                                    AttributesMask=mask,
                                                    Caller=GUID_ZERO)
        io = DeviceIoControl(self._path, True)
        io.ioctl_disk_set_disk_attributes(set_struct)

    def online(self):
        self._set_disk_attrttributes(0, constants.DISK_ATTRIBUTE_OFFLINE)

    def offline(self):
        self._set_disk_attrttributes(constants.DISK_ATTRIBUTE_OFFLINE, constants.DISK_ATTRIBUTE_OFFLINE)

    def read_only(self):
        self._set_disk_attrttributes(constants.DISK_ATTRIBUTE_READ_ONLY, constants.DISK_ATTRIBUTE_READ_ONLY)

    def read_write(self):
        self._set_disk_attrttributes(0, constants.DISK_ATTRIBUTE_READ_ONLY)

    def get_volume_number(self):
        self._io.ioctl_volume_query_volume_number()




PARTITION_STYLE_MBR = 0
PARTITION_STYLE_GPT = 1
PARTITION_STYLE_RAW = 2

PARTITION_ENTRY_UNUSED = 0x00
PARTITION_FAT_12 = 0x01
PARTITION_XENIX_1 = 0x02
PARTITION_XENIX_2 = 0x03
PARTITION_FAT_16 = 0x04
PARTITION_EXTENDED = 0x05
PARTITION_HUGE = 0x06
PARTITION_IFS = 0x07
PARTITION_OS2BOOTMGR = 0x0A
PARTITION_FAT32 = 0x0B
PARTITION_FAT32_XINT13 = 0x0C
PARTITION_XINT13 = 0x0E
PARTITION_XINT13_EXTENDED = 0x0F
PARTITION_PREP = 0x41
PARTITION_LDM = 0x42
PARTITION_UNIX = 0x63

TRUE = 1

DISK_ATTRIBUTE_OFFLINE = 0x0000000000000001
DISK_ATTRIBUTE_READ_ONLY = 0x0000000000000002

DISK_SAN_POLICY_UNKNOWN = 0
DISK_SAN_POLICY_ONLINE = 1
DISK_SAN_POLICY_OFFLINE_SHARED = 2
DISK_SAN_POLICY_OFFLINE = 3
DISK_SAN_POLICY_MAX = 4

MOUNTMGR_AUTO_MOUNT_STATE_DISABLED = 0
MOUNTMGR_AUTO_MOUNT_STATE_ENABLED = 1





from infi import unittest
import os
from ... import wmi

class WmiTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(WmiTestCase, cls).setUpClass()
        if os.name != 'nt':
            raise unittest.SkipTest("This test case should run only on Windows")

    def test_get_drives(self):
        client = wmi.WmiClient()
        drives = wmi.get_disk_drives(client)
        self.assertGreater(len(drives), 0)

    def test_get_partitions(self):
        client = wmi.WmiClient()
        drives = wmi.get_disk_drives(client)
        self.assertGreater(len(drives), 0)
        partitions_by_drive = {drive.Name:wmi.get_paritions_of_disk_drive(client, drive) for drive in drives.values() }
        self.assertGreater(len(partitions_by_drive), 0)
        self.assertGreater(len(partitions_by_drive[ur"\\.\PHYSICALDRIVE0"]), 0)



from infi import unittest
from ..mount_manager import MountManager
import ctypes

class MockVolume(object):
    def __init__(self):
        super(MockVolume, self).__init__()
        self._path = u'\\\\?\\GLOBALROOT\\Device\\HarddiskVolume35'

class InputBufferTestCase(unittest.TestCase):
    def test_create_input_buffer(self):
        size = ctypes.sizeof(ctypes.c_wchar)
        m = MountManager()
        buffer = m._create_input_buffer_for_query_points_ioctl(MockVolume())
        self.assertEqual(buffer.raw[0 - size - 1:], '\x00' * (size + 1))



import unittest
import ctypes

# pylint: disable=W0614,R0904,W0401,W0622
from .structures import *
from .constants import PARTITION_STYLE_MBR, PARTITION_STYLE_GPT

def _sizeof(struct):
    return struct.min_max_sizeof().max

class CreateDiskTestCase(unittest.TestCase):
    def _get_mbr(self):
        struct = CREATE_DISK_MBR()
        struct.Signature = 1
        return struct

    def test_create_disk__mbr(self):
        struct = self._get_mbr()
        CREATE_DISK_MBR.write_to_string(struct)

    def _get_guid(self):
        struct = GUID.create_from_string('1' * 16)
        return struct

    def test_guid(self):
        struct = self._get_guid()
        GUID.write_to_string(struct)

    def _get_gpt(self):
        struct = CREATE_DISK_GPT()
        struct.DiskId = self._get_guid()
        struct.MaxPartitionCount = 1
        return struct

    def test_create_disk__gpt(self):
        struct = self._get_gpt()
        CREATE_DISK_GPT.write_to_string(struct)

    def test_create_disk(self):
        struct = CREATE_DISK()
        struct.PartitionStyle = 1
        struct.union = self._get_mbr()
        CREATE_DISK.write_to_string(struct)

class StructuresTestCase(unittest.TestCase):
    def test_partition_information__mbr(self):
        buffer = ctypes.c_buffer("\x00" * 137, 137)
        struct = PARTITION_INFORMATION_EX.create_from_string(buffer)

    def test_create_disk__mbr(self):
        """:param partition_style: int
        :param signature: int
        :param disk_id: 16-byte string
        :param max_partition_count: int"""
        struct = CREATE_DISK()
        struct.PartitionStyle = 1
        if struct.PartitionStyle == PARTITION_STYLE_MBR:
            klass = CREATE_DISK_MBR
            union = klass()
            union.Signature = 1
        else:
            klass = CREATE_DISK_GPT
            union = klass()
            union.DiskId = GUID.create_from_string("\x00"*16)
            union.MaxPartitionCount = 128
        struct.union = union
        size = struct.min_max_sizeof().max
        buffer = ctypes.c_buffer(CREATE_DISK.write_to_string(struct), size)

    def test_sizes(self):
        self.assertEqual(_sizeof(DRIVE_LAYOUT_INFORMATION_GPT), 40)
        self.assertEqual(_sizeof(DRIVE_LAYOUT_INFORMATION_MBR), 40)
        self.assertEqual(_sizeof(PARTITION_INFORMATION_GPT), 112)
        self.assertEqual(_sizeof(PARTITION_INFORMATION_MBR), 112)
        self.assertEqual(_sizeof(PARTITION_INFORMATION_EX), 144)
        self.assertEqual(_sizeof(DRIVE_LAYOUT_INFORMATION_EX), 48)




__import__("pkg_resources").declare_namespace(__name__)

import re
from infi.wmi import WmiClient, WmiObject
from .model import DiskDrive, DiskPartition, DISKDRIVES_QUERY, DISKPARTITION_QUERY
from .model import DISKDRIVE_TO_DISKPARTITIONS_QUERY

def escape_string(string):
    return re.sub("[\\\\'\"]", '\\\\\\g<0>', string)

def get_disk_drives(wmi_client):
    """:returns: a dictionary of (DiskDrive.Name, DiskDrive) items"""
    devices = dict()
    for result in wmi_client.execute_query(DISKDRIVES_QUERY):
        device = DiskDrive(result)
        devices[device.Name] = device
    return devices

def get_paritions_of_disk_drive(wmi_client, disk_drive):
    """:returns: a list of DiskPartition objects"""
    mapping_query = DISKDRIVE_TO_DISKPARTITIONS_QUERY.format(escape_string(disk_drive.get_path()))
    devices = []
    for mapping in wmi_client.execute_query(mapping_query):
        _, device_id = WmiObject(mapping).get_wmi_attribute("Dependent").split("=")
        partition_query = DISKPARTITION_QUERY.format(device_id)
        for result in wmi_client.execute_query(partition_query):
            devices.append(DiskPartition(result))
    return devices

def iter_volumes(wmi_client):
    from .model import Volume, VOLUME_QUERY
    for item in wmi_client.execute_query(VOLUME_QUERY):
        yield Volume(item)




from ..ioctl import DeviceIoControl, generate_signature, generate_guid, structures, constants
from ..ioctl.constants import PARTITION_STYLE_MBR, PARTITION_STYLE_GPT, PARTITION_STYLE_RAW
from ..ioctl.constants import MOUNTMGR_AUTO_MOUNT_STATE_DISABLED, MOUNTMGR_AUTO_MOUNT_STATE_ENABLED
from infi.pyutils.lazy import cached_method
from infi.diskmanagement.ioctl import GUID_ZERO, _sizeof
from infi.instruct import Struct, BitFields, BitField, BitPadding, ULInt32
from string import ascii_uppercase
import ctypes

MAX_PATH_NAMES = 32767

class AVAILABLE_DRIVE_LETTERS(Struct):
    _fields_ = BitFields(*([BitField(letter, 1) for letter in ascii_uppercase] + [BitPadding(6), ]))

def _slice_unicode_string_from_buffer(buffer, offset, length):
    return ctypes.wstring_at(ctypes.addressof(buffer) + offset, length / ctypes.sizeof(ctypes.c_wchar))

def _rstrip(string):
    return string.rstrip(u'\\')

def _get_unicode_buffer(size=MAX_PATH_NAMES):
    buffer = create_unicode_buffer(size)
    return buffer, size

class MountManager(object):
    def __init__(self):
        super(MountManager, self).__init__()
        self._path = r"\\.\MountPointManager"
        self._io = DeviceIoControl(self._path, True)

    def __repr__(self):
        return "MountManager <{}>".format(self._path)

    def get_avaialable_drive_letters(self):
        bitmask = ctypes.windll.kernel32.GetLogicalDrives()
        struct = AVAILABLE_DRIVE_LETTERS.create_from_string(ULInt32.write_to_string(bitmask))
        return [u"{}:\\".format(letter) for letter in filter(lambda letter: getattr(struct, letter) == 0,
                                             [field.name for field in struct._fields_.fields[2:26]])]

    def _create_input_buffer_for_query_points_ioctl(self, volume):
        volume_guid = getattr(volume, "_path", volume).split('\\')[-1]
        device_name = r"\Device\{}".format(volume_guid)
        unicode_buffer = ctypes.create_unicode_buffer(device_name)
        from os.path import sep
        buffer_string = ctypes.string_at(ctypes.addressof(unicode_buffer),
                                         len(unicode_buffer) * ctypes.sizeof(ctypes.c_wchar))
        triplet = structures.MOUNTMGR_MOUNT_POINT(SymbolicLinkNameOffset=0, SymbolicLinkNameLength=0,
                                                  UniqueIdOffset=0, UniqueIdLength=0,
                                                  DeviceNameOffset=_sizeof(structures.MOUNTMGR_MOUNT_POINT),
                                                  DeviceNameLength=len(buffer_string) - ctypes.sizeof(ctypes.c_wchar))
        triplet_string = structures.MOUNTMGR_MOUNT_POINT.write_to_string(triplet)
        input_buffer = ctypes.c_buffer(triplet_string + buffer_string, len(triplet_string) + len(buffer_string))
        return input_buffer

    def _iter_volume_mount_points(self, volume):
        input_buffer = self._create_input_buffer_for_query_points_ioctl(volume)
        output_buffer = self._io.ioctl_mountmgr_query_points(input_buffer, len(input_buffer) + 1)
        struct = structures.MOUNTMGR_MOUNT_POINTS.create_from_string(output_buffer)
        for item in struct.MountPoints:
            offset, length = item.SymbolicLinkNameOffset, item.SymbolicLinkNameLength
            yield _slice_unicode_string_from_buffer(output_buffer, offset, length)

    def get_volume_drive_letter(self, volume):
        for item in self._iter_volume_mount_points(volume):
            if "DosDevices\\" in item:
                return item.split('\\')[-1][0]

    def get_volume_mount_points(self, volume_guid):
        if not volume_guid.endswith('\\'):
            volume_guid = u"{}\\".format(volume_guid)
        volumePathNames = create_unicode_buffer(MAX_PATH_NAMES)
        returnLength = DWORD(0)
        GetVolumePathNamesForVolumeNameW(volumeName=volume_guid, volumePathNames=volumePathNames,
                                         returnLength=returnLength)
        return filter(lambda string: string != u'',
                      ctypes.wstring_at(ctypes.addressof(volumePathNames), returnLength.value).split(u"\x00"))

    def add_volume_mount_point(self, volume_guid, mount_point):
        if not volume_guid.endswith('\\'):
            volume_guid = u"{}\\".format(volume_guid)
        SetVolumeMountPointW(create_unicode_buffer(mount_point), create_unicode_buffer(volume_guid))

    def remove_mount_point(self, volume_guid, mount_point):
        if not mount_point.endswith('\\'):
            mount_point = u"{}\\".format(mount_point)
        mount_points = self.get_volume_mount_points(volume_guid)
        assert mount_point in mount_points, "{} is not a mount point".format(mount_point)
        DeleteVolumeMountPointW(create_unicode_buffer(mount_point))

    def get_volume_label(self, mount_point):
        if not mount_point.endswith('\\'):
            mount_point = u"{}\\".format(mount_point)
        volume_name = create_unicode_buffer(MAX_PATH_NAMES)
        volume_name_size = DWORD(MAX_PATH_NAMES)
        GetVolumeInformationW(rootPathName=mount_point,
                              volumeNameBuffer=volume_name, volumeNameSize=volume_name_size,
                              volumeSerialNumber=None, maximumComponentLength=None, fileSystemFlags=None,
                              fileSystemNameBuffer=None, fileSystemNameSize=DWORD(0))
        return ctypes.wstring_at(ctypes.addressof(volume_name))

    def set_volume_label(self, mount_point, label):
        """:raises: `OverflowError if the label is too big"""
        if not mount_point.endswith('\\'):
            mount_point = u"{}\\".format(mount_point)
        try:
            SetVolumeLabelW(rootPathName=create_unicode_buffer(mount_point),
                            volumeName=create_unicode_buffer(label))
        except WindowsException, e:
            raise

    def is_auto_mount(self):
        return self._io.ioctl_mountmgr_query_auto_mount() == MOUNTMGR_AUTO_MOUNT_STATE_ENABLED

    def enable_auto_mount(self):
        return self._io.ioctl_mountmgr_set_auto_mount(MOUNTMGR_AUTO_MOUNT_STATE_ENABLED)

    def disable_auto_mount(self):
        return self._io.ioctl_mountmgr_set_auto_mount(MOUNTMGR_AUTO_MOUNT_STATE_DISABLED)

    @cached_method
    def get_volume_extents(self, volume_guid):
        return DeviceIoControl(volume_guid, False).ioctl_volume_get_volume_disk_extents().Extents

    @cached_method
    def get_volume_guids(self):
        return list(self.iter_volume_guids())

    def iter_volume_guids(self):
        buffer, length = _get_unicode_buffer()
        search_handle = None
        try:
            search_handle = FindFirstVolumeW(buffer, length)
            while True:
                if buffer.value:
                    yield buffer.value
                FindNextVolumeW(search_handle, buffer, length)
        except WindowsException, e:
            if e.winerror != ERROR_NO_MORE_FILES:
                raise
        finally:
            if search_handle is not None:
                FindVolumeClose(search_handle)

    def get_mounts_of_all_volumes(self):
        """:returns: a dictionary with volume GUIDs as keys and list of mount points as values"""
        return {_rstrip(volume_guid): self.get_volume_mount_points(volume_guid) \
                for volume_guid in self.iter_volume_guids()}

class PartitionManager(object):
    def __init__(self):
        super(PartitionManager, self).__init__()
        self._path = r"\\.\PartmgrControl"
        self._io = DeviceIoControl(self._path, True)

    def __repr__(self):
        return "PartitionManager <{}>".format(self._path)

    def get_san_policy(self):
        return self._io.ioctl_disk_get_san_settings().SanPolicy

    def set_san_policy(self, san_policy):
        version = self._io.ioctl_disk_get_san_settings().Version
        settings = structures.DISK_SAN_SETTINGS(Version=version, SanPolicy=san_policy)
        return self._io.ioctl_disk_set_san_settings(settings)

import infi.wioctl
from infi.cwrap import WrappedFunction, IN, IN_OUT, OUT
from ctypes import c_wchar_p as LPCWSTR
from ctypes import c_wchar_p as LPWSTR
from ctypes import c_ulong as DWORD
from ctypes import POINTER, create_unicode_buffer

class GetVolumePathNamesForVolumeNameW(WrappedFunction):
    return_value = infi.wioctl.api.BOOL

    @classmethod
    def get_errcheck(cls):
        return infi.wioctl.api.errcheck_bool()

    @classmethod
    def get_library_name(cls):
        return 'kernel32'

    @classmethod
    def get_parameters(cls):
        return ((LPCWSTR, IN, "volumeName"),
                (LPWSTR, IN_OUT, "volumePathNames"),
                (DWORD, IN, "bufferLength", DWORD(MAX_PATH_NAMES)),
                (POINTER(DWORD), IN_OUT, "returnLength"))

class SetVolumeMountPointW(WrappedFunction):
    return_value = infi.wioctl.api.BOOL

    @classmethod
    def get_errcheck(cls):
        return infi.wioctl.api.errcheck_bool()

    @classmethod
    def get_library_name(cls):
        return 'kernel32'

    @classmethod
    def get_parameters(cls):
        return ((LPCWSTR, IN, "volumeMountPoint"),
                (LPCWSTR, IN, "volumeName"))

class DeleteVolumeMountPointW(WrappedFunction):
    return_value = infi.wioctl.api.BOOL

    @classmethod
    def get_errcheck(cls):
        return infi.wioctl.api.errcheck_bool()

    @classmethod
    def get_library_name(cls):
        return 'kernel32'

    @classmethod
    def get_parameters(cls):
        return ((LPCWSTR, IN, "volumeMountPoint"),)

from infi.wioctl.api import HANDLE, errcheck_invalid_handle, WindowsException

class FindFirstVolumeW(WrappedFunction):
    return_value = HANDLE

    @classmethod
    def get_errcheck(cls):
        return errcheck_invalid_handle()

    @classmethod
    def get_library_name(cls):
        return 'kernel32'

    @classmethod
    def get_parameters(cls):
        return ((LPCWSTR, IN_OUT, "volumeName"),
                (DWORD, IN, "bufferLength"))

class FindVolumeClose(WrappedFunction):
    return_value = HANDLE

    @classmethod
    def get_errcheck(cls):
        return errcheck_invalid_handle()

    @classmethod
    def get_library_name(cls):
        return 'kernel32'

    @classmethod
    def get_parameters(cls):
        return ((HANDLE, IN, "findVolume"),)

class FindNextVolumeW(WrappedFunction):
    return_value = infi.wioctl.api.BOOL

    @classmethod
    def get_errcheck(cls):
        return infi.wioctl.api.errcheck_bool()

    @classmethod
    def get_library_name(cls):
        return 'kernel32'

    @classmethod
    def get_parameters(cls):
        return ((HANDLE, IN, "findVolume"),
                (LPWSTR, IN_OUT, "volumeName"),
                (DWORD, IN, "bufferLength"))

ERROR_NO_MORE_FILES = 18

class FindFirstVolumeMountPointW(WrappedFunction):
    return_value = HANDLE

    @classmethod
    def get_errcheck(cls):
        return errcheck_invalid_handle()

    @classmethod
    def get_library_name(cls):
        return 'kernel32'

    @classmethod
    def get_parameters(cls):
        return ((LPWSTR, IN, "rootPathName"),
                (LPWSTR, IN_OUT, "volumeMountPoint"),
                (DWORD, IN, "bufferLength"))

class FindVolumeMountPointClose(WrappedFunction):
    return_value = HANDLE

    @classmethod
    def get_errcheck(cls):
        return errcheck_invalid_handle()

    @classmethod
    def get_library_name(cls):
        return 'kernel32'

    @classmethod
    def get_parameters(cls):
        return ((HANDLE, IN, "findVolumeMountPoint"),)

class FindNextVolumeMountPointW(WrappedFunction):
    return_value = infi.wioctl.api.BOOL

    @classmethod
    def get_errcheck(cls):
        return infi.wioctl.api.errcheck_bool()

    @classmethod
    def get_library_name(cls):
        return 'kernel32'

    @classmethod
    def get_parameters(cls):
        return ((HANDLE, IN, "findVolumeMountPoint"),
                (LPWSTR, IN_OUT, "volumeMountPoint"),
                (DWORD, IN, "bufferLength"))

class GetVolumeInformationW(WrappedFunction):
    return_value = infi.wioctl.api.BOOL

    @classmethod
    def get_errcheck(cls):
        return errcheck_invalid_handle()

    @classmethod
    def get_library_name(cls):
        return 'kernel32'

    @classmethod
    def get_parameters(cls):
        return ((LPWSTR, IN, "rootPathName"),
                (LPWSTR, IN_OUT, "volumeNameBuffer"),
                (DWORD, IN, "volumeNameSize"),
                (POINTER(DWORD), IN_OUT, "volumeSerialNumber"),
                (POINTER(DWORD), IN_OUT, "maximumComponentLength"),
                (POINTER(DWORD), IN_OUT, "fileSystemFlags"),
                (LPWSTR, IN_OUT, "fileSystemNameBuffer"),
                (DWORD, IN, "fileSystemNameSize"))


class SetVolumeLabelW(WrappedFunction):
    return_value = infi.wioctl.api.BOOL

    @classmethod
    def get_errcheck(cls):
        return infi.wioctl.api.errcheck_bool()

    @classmethod
    def get_library_name(cls):
        return 'kernel32'

    @classmethod
    def get_parameters(cls):
        return ((HANDLE, IN, "rootPathName"),
                (LPWSTR, IN, "volumeName"),)




from .. import WmiObject
from infi.pyutils.lazy import cached_method

DISKDRIVES_QUERY = "SELECT * FROM Win32_DiskDrive"
VOLUME_QUERY = "SELECT * FROM Win32_Volume"
DISKDRIVE_TO_DISKPARTITIONS_QUERY = r'SELECT * FROM Win32_DiskDriveToDiskPartition WHERE Antecedent="{}"'
DISKPARTITION_QUERY = r'SELECT * FROM Win32_DiskPartition WHERE DeviceID={}'

class DiskPartition(WmiObject):
    @property
    def Name(self):
        return self.get_wmi_attribute("Name")

    def __repr__(self):
        return "DiskPartition <{}>".format(self.Name)

class DiskDrive(WmiObject):
    @property
    def Name(self):
        return self.get_wmi_attribute("Name")

    @property
    def SerialNumber(self):
        return self.get_wmi_attribute("SerialNumber")

    def __repr__(self):
        return "DiskDrive <{}>".format(self.Name)

class Volume(WmiObject):
    @property
    def DeviceID(self):
        return self.get_wmi_attribute("DeviceID")

    def Format(self, ClusterSize=0, EnableCompression=False, FileSystem="NTFS", QuickFormat=True):
        method = self._object.Methods_("Format")
        parameters = method.InParameters.SpawnInstance_()
        parameters.Properties_.Item("ClusterSize").Value = ClusterSize
        parameters.Properties_.Item("EnableCompression").Value = EnableCompression
        parameters.Properties_.Item("FileSystem").Value = FileSystem
        parameters.Properties_.Item("QuickFormat").Value = QuickFormat
        _ = self._object.ExecMethod_(method.Name, parameters)




__version__ = "0.3.9"
__git_commiter_name__ = "Arnon Yaari"
__git_commiter_email__ = "arnony@infinidat.com"
__git_branch__ = '(detached from v0.3.9)'
__git_remote_tracking_branch__ = '(No remote tracking)'
__git_remote_url__ = '(Not remote tracking)'
__git_head_hash__ = '69e422900e88db0c1ce10ee38ccd03b5bb34338f'
__git_head_subject__ = 'Finished release v0.3.9.'
__git_head_message__ = ''
__git_dirty_diff__ = ''



SETUP_INFO = dict(
    name = 'infi.diskmanagement',
    version = '0.3.9',
    author = 'Arnon Yaari',
    author_email = 'arnony@infinidat.com',

    url = 'https://github.com/Infinidat/infi.diskmanagement',
    license = 'PSF',
    description = """Windows Disk Management wrapping in Python""",
    long_description = """This module gives the same functionality as diskpart. But unlike diskpart, it does not use VDS, it uses SetupAPI and direct IOCTLs to the disks, volumes, and the mount and partitions managers""",

    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers = [
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Python Software Foundation License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],

    install_requires = ['capacity',
'infi.cwrap',
'infi.devicemanager',
'infi.instruct',
'infi.pyutils',
'infi.wioctl>=0.1.3',
'infi.wmi',
'setuptools',
'waiting'],
    namespace_packages = ['infi'],

    package_dir = {'': 'src'},
    package_data = {'': []},
    include_package_data = True,
    zip_safe = False,

    entry_points = dict(
        console_scripts = [],
        gui_scripts = [],
        ),
)

if SETUP_INFO['url'] is None:
    _ = SETUP_INFO.pop('url')

def setup():
    from setuptools import setup as _setup
    from setuptools import find_packages
    SETUP_INFO['packages'] = find_packages('src')
    _setup(**SETUP_INFO)

if __name__ == '__main__':
    setup()

